//@author: elliotwoods, vvvv group
//@help: shadow casting shadow (based on GourandPoint)
//@tags: shadow, shading, blinn
//@credits:

// --------------------------------------------------------------------------------------------------
// PARAMETERS:
// --------------------------------------------------------------------------------------------------

//transforms
float4x4 tW: WORLD;        //the models world matrix
float4x4 tV: VIEW;         //view matrix as set via Renderer (DX9)
float4x4 tWV: WORLDVIEW;
float4x4 tWVP: WORLDVIEWPROJECTION;
float4x4 tP: PROJECTION;   //projection matrix as set via Renderer (DX9)

//light properties
float3 lPos <string uiname="Light Position";> = {0, 5, -2};       //light position in world space
float lAtt0 <String uiname="Light Attenuation 0"; float uimin=0.0;> = 0;
float lAtt1 <String uiname="Light Attenuation 1"; float uimin=0.0;> = 0.3;
float lAtt2 <String uiname="Light Attenuation 2"; float uimin=0.0;> = 0;
float4 lAmb  : COLOR <String uiname="Ambient Color";>  = {0.15, 0.15, 0.15, 1};
float4 lDiff : COLOR <String uiname="Diffuse Color";>  = {0.85, 0.85, 0.85, 1};
float4 lSpec : COLOR <String uiname="Specular Color";> = {0.35, 0.35, 0.35, 1};
float lPower <String uiname="Power"; float uimin=0.0;> = 25.0;     //shininess of specular highlight
float lRange <String uiname="Light Range"; float uimin=0.0;> = 10.0;

float Alpha <float uimin=0.0; float uimax=1.0;> = 1;

//texture
texture Tex <string uiname="Texture";>;
sampler Samp = sampler_state    //sampler for doing the texture-lookup
{
    Texture   = (Tex);          //apply a texture to the sampler
    MipFilter = LINEAR;         //set the sampler states
    MinFilter = LINEAR;
    MagFilter = LINEAR;
};

texture TexShadow <string uiname="Shadow Map";>;
sampler SampShadow = sampler_state    //sampler for doing the texture-lookup
{
    Texture   = (TexShadow);          //apply a texture to the sampler
    MipFilter = NONE;         //set the sampler states
    MinFilter = NONE;
    MagFilter = NONE;
};

float4x4 tShadow <string uiname="Shadow ViewProjection";>;

float4x4 tTex: TEXTUREMATRIX <string uiname="Texture Transform";>;
float4x4 tColor <string uiname="Color Transform";>;

struct vs2ps
{
    float4 PosWVP: POSITION;
    float4 TexCd : TEXCOORD0;
	float4 ShadowCd : TEXCOORD1;
    float4 Diffuse : COLOR0;
    float4 Specular : COLOR1;
};

// --------------------------------------------------------------------------------------------------
// VERTEXSHADERS
// --------------------------------------------------------------------------------------------------

vs2ps VS(
    float4 PosO: POSITION,
    float3 NormO: NORMAL,
    float4 TexCd : TEXCOORD0)
{
    //inititalize all fields of output struct with 0
    vs2ps Out = (vs2ps)0;

    //distance from light to vertex
    float4 PosW = mul(PosO, tW);
    float d = distance(PosW, lPos);

    float atten = 0;

    //compute attenuation only if vertex within lightrange
    if (d<lRange)
    {
       atten = 1/(saturate(lAtt0) + saturate(lAtt1) * d + saturate(lAtt2) * pow(d, 2));
    }
    
    //inverse light direction in view space
    float3 LightDirW = normalize(lPos - PosW);
    float3 LightDirV = mul(LightDirW, tV);
    
    //normal in view space
    float3 NormV = normalize(mul(NormO, tWV));

    //view direction = inverse vertexposition in viewspace
    float3 PosV = mul(PosW, tV);
    float3 ViewDirV = normalize(-PosV);
    //halfvector
    float3 H = normalize(ViewDirV + LightDirV);

    //compute blinn lighting
    float4 shades = lit(dot(NormV, LightDirV), dot(NormV, H), lPower);

    float4 diff = lDiff * shades.y * atten;
    diff.a = 1;
    float4 spec = lSpec * shades.z * atten;
    spec.a = 1;

    //position (projected)
    Out.PosWVP  = mul(PosO, tWVP);
    Out.TexCd = mul(TexCd, tTex);
    Out.Diffuse = diff;
    Out.Specular = spec;
	
	Out.ShadowCd = mul(PosW, tShadow);

    return Out;
}

// --------------------------------------------------------------------------------------------------
// PIXELSHADERS:
// --------------------------------------------------------------------------------------------------

float4 PS(vs2ps In): COLOR
{
    //calculate coordinates in shadow space
	In.ShadowCd /= In.ShadowCd.w;
	float2 ShadowLookup = In.ShadowCd;
	ShadowLookup.y *= -1.0f;
	ShadowLookup += 1.0f;
	ShadowLookup /= 2.0f;
	
	//read off calculated depth and rendered depth
	float3 depthLookup = tex2D(SampShadow, ShadowLookup).rgb;
	//This is a strange feature of the depth buffer
	//it seems that the depth buffer reports the square root 
	//of the value, so we must square it to retreive the
	//true depth value.
	float depth = pow(depthLookup.z, 2);
	float z = In.ShadowCd.z;
	
	//check if depths match (to an epsilon)
	bool hit = (z - depth) < 0.0000005;
	
	//kill light outside of view frustum
	float inside = ShadowLookup.x > 0 && ShadowLookup.x < 1.0f;
	inside = inside && ShadowLookup.y > 0 && ShadowLookup.y < 1.0f;
	
	//fade of light at edges (flashlight effect)
	inside *= pow(1.0f - smoothstep(0, 1, length(In.ShadowCd.xy)), 0.2);
	
	//sum up lighting
	float4 Light = In.Diffuse + In.Specular;
	Light *= inside * hit;
	Light += lAmb;
	Light.a = 1.0f;
	
    float4 col = tex2D(Samp, In.TexCd);	
    col.rgb *= Light;
    col = mul(col, tColor);	
    col.a *= Alpha;
	
    return col;
}


// --------------------------------------------------------------------------------------------------
// TECHNIQUES:
// --------------------------------------------------------------------------------------------------

technique TShadowCast
{
    pass P0
    {
        //Wrap0 = U;  // useful when mesh is round like a sphere
        VertexShader = compile vs_1_0 VS();
        PixelShader = compile ps_2_0 PS();
    }
}